import sqlite3
from six import string_types
from operator import attrgetter
from cachetools import LRUCache, cachedmethod

from dark.database import getDatabaseConnection


class LineageFetcher(object):
    """
    Provide access to the NCBI taxonomy database so we can retrieve the lineage
    of title sequences hit by BLAST.
    """
    def __init__(self, db=None, cursor=None):
        self._db = db or getDatabaseConnection()
        self._cursor = cursor or self._db.cursor()
        self._cache = {}

    def lineage(self, title):
        """
        Get lineage information from the taxonomy database for a given title.

        @param title: A C{str} sequence title (e.g., from a BLAST hit). Of the
            form 'gi|63148399|gb|DQ011818.1| Description...'. It is the gi
            number (63148399 in this example) that is looked up in the taxonomy
            database.
        @return: A C{list} of the taxonomic categories of the title. Each list
            element is an (C{int}, C{str}) 2-tuple, giving a taxonomy id and
            a scientific name. The first element in the list will correspond to
            C{title}, and each successive element is the parent of the
            preceeding one. If no taxonomy is found, the returned list will be
            empty.
        """
        if title in self._cache:
            return self._cache[title]

        lineage = []
        gi = int(title.split('|')[1])
        query = 'SELECT taxID FROM gi_taxid WHERE gi = %d' % gi

        try:
            while True:
                self._cursor.execute(query)
                taxID = self._cursor.fetchone()[0]
                if taxID == 1:
                    break
                query = 'SELECT name FROM names WHERE taxId = %s' % taxID
                self._cursor.execute(query)
                scientificName = self._cursor.fetchone()[0]
                lineage.append((taxID, scientificName))
                # Move up to the parent.
                query = ('SELECT parent_taxID FROM nodes WHERE taxID = %s' %
                         taxID)
        except TypeError:
            lineage = []

        self._cache[title] = lineage
        return lineage

    def close(self):
        """
        Close the database connection and render self invalid. Any subsequent
        re-use of self will raise an error.
        """
        self._cursor.close()
        self._db.close()
        self._cursor = self._db = self._cache = None


class AccessionLineageFetcher(object):
    """
    Provide access to the NCBI taxonomy database so we can retrieve the
    taxonomy lineage corresponding to an accession number.

    @param dbFilenameOrConnection: Either a C{str} database filename or an
        open sqlite3 database. The database must contain taxonomy information
        with tables and columns named as in the building scripts used in
        https://github.com/acorg/ncbi-taxonomy-database
    """
    CACHE_SIZE = 10E6

    def __init__(self, dbFilenameOrConnection):
        if isinstance(dbFilenameOrConnection, string_types):
            self._db = sqlite3.connect(dbFilenameOrConnection)
            self._closeConnection = True
        else:
            self._db = dbFilenameOrConnection
            self._closeConnection = False
        self._cache = LRUCache(maxsize=self.CACHE_SIZE)

    def lineageFromTaxid(self, taxid):
        cursor = self._db.cursor()
        execute = cursor.execute
        fetchone = cursor.fetchone

        lineage = []

        while taxid != 1:
            execute('SELECT name FROM names WHERE taxid = ?', (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    'Could not find taxonomy id %r in names table' %
                    (taxid,))
            name = result[0]
            execute('SELECT parent_taxid, rank FROM nodes WHERE taxid = ?',
                    (taxid,))
            result = fetchone()
            if result is None:
                raise ValueError(
                    'Could not find taxonomy id %r in nodes table' %
                    (taxid,))
            parentTaxid, rank = result
            lineage.append((taxid, name, rank))
            taxid = parentTaxid

        return tuple(lineage) or None

    @cachedmethod(attrgetter('_cache'))
    def lineage(self, accession):
        """
        Get lineage information from the taxonomy database for an
        accession number or taxonomy id.

        @param accession: A C{str} accession number or C{int} taxonomy id.
            If a C{str}, C{accession} must include the version (e.g., the
            ".1" part of "DQ011818.1") if the taxonomy database in use uses
            it (as is the case with the database built by
            https://github.com/acorg/ncbi-taxonomy-database
        @raise ValueError: If a taxonomy id cannot be found in the names or
            nodes table.
        @raise AttributeError: If C{self.close} has been called.
        @return: A C{tuple} of the taxonomic categories of the title. Each
            tuple element is a 3-tuple of (C{int}, C{str}, C{str}) giving a
            taxonomy id a (scientific) name, and the rank (species, genus,
            etc). The first element in the list corresponds to the passed
            C{accession}, number and each successive element is the parent of
            the preceeding one. The top level of the taxonomy hierarchy
            (with taxid = 1) is not included in the returned list. If no
            taxonomy information is found for C{accession}, return C{None}.
        """
        if isinstance(accession, int):
            return self.lineageFromTaxid(accession)
        else:
            cursor = self._db.cursor()
            fetchone = cursor.fetchone

            cursor.execute(
                'SELECT taxid FROM accession_taxid WHERE accession = ?',
                (accession,))
            row = fetchone()
            if row:
                return self.lineageFromTaxid(int(row[0]))

    def close(self):
        """
        Close the database connection (if we opened it).
        """
        if self._closeConnection:
            self._db.close()
        # Set self._db to None so self.lineage will raise an AttributeError
        # exception if it is called again.
        self._db = None

    def __enter__(self):
        return self

    def __exit__(self, excType, excValue, traceback):
        self.close()


def isRetrovirus(lineage):
    """
    Determine whether a lineage corresponds to a retrovirus.

    @param lineage: A C{tuple} of taxonomy id, scientific name, and rank
        as returned by C{AccessionLineageFetcher.lineage}.
    @return: C{True} if the lineage corresponds to a retrovirus, C{False}
        otherwise.
    """
    for taxid, name, rank in lineage:
        if rank == 'family' and name == 'Retroviridae':
            return True

    return False


def isRNAVirus(lineage):
    """
    Determine whether a lineage corresponds to an RNA virus.

    @param lineage: A C{tuple} of taxonomy id, scientific name, and rank
        as returned by C{AccessionLineageFetcher.lineage}.
    @return: C{True} if the lineage corresponds to an RNA virus, C{False}
        otherwise.
    """
    for taxid, name, rank in lineage:
        if rank == 'realm' and name == 'Riboviria':
            return True

    return False


def _preprocessLineage(lineage):
    """
    Pre-process a lineage to make it easier to deal with by others.

    @param lineage: A C{tuple} of taxonomy id, scientific name, and rank
        as returned by C{AccessionLineageFetcher.lineage}.
    @return: A 3-tuple of taxids, names, and ranks (all as C{str}ings).
    """
    taxids, names, ranks = [], [], []

    for (taxid, name, rank) in lineage:
        taxids.append(str(taxid))
        ranks.append('-' if rank == 'no rank' else rank)
        names.append(name)

    return taxids, names, ranks


def formatLineage(lineage, namesOnly=False, separator=None, prefix=''):
    """
    Format a lineage for printing.

    @param lineage: A C{tuple} of taxonomy id, scientific name, and rank
        as returned by C{AccessionLineageFetcher.lineage}.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: A C{str} separator to put between fields. If C{None},
        return a space-padded aligned columns.
    @param prefix: A C{str} to put at the start of each line.
    @return: A formatted C{str} for printing.
    """
    taxids, names, ranks = _preprocessLineage(lineage)

    if namesOnly:
        # The separator is guaranteed to be set by our caller.
        return prefix + separator.join(names)

    if separator is not None:
        return '\n'.join(
            '%s%s%s%s%s%s' % (prefix, rank, separator, name, separator, taxid)
            for (taxid, name, rank) in zip(taxids, names, ranks))
    else:
        taxidWidth = max(len(x) for x in taxids)
        nameWidth = max(len(x) for x in names)
        rankWidth = max(len(x) for x in ranks)

        return '\n'.join(
            '%s%-*s %-*s %*s' % (
                prefix, rankWidth, rank, nameWidth, name, taxidWidth, taxid)
            for (taxid, name, rank) in zip(taxids, names, ranks))


def lineageTaxonomyLinks(lineage):
    """
    Get HTML links for a lineage.

    @param lineage: A C{tuple} of taxonomy id, scientific name, and rank
        as returned by C{AccessionLineageFetcher.lineage}.
    @return: A C{list} of HTML C{str} links.
    """
    taxids, names, _ = _preprocessLineage(lineage)

    assert names[-1] == 'Viruses'
    names[0] = 'taxon'

    return [
        '<a href="%s%s">%s</a>' % (
            'https://www.ncbi.nlm.nih.gov/Taxonomy/Browser/wwwtax.cgi?id=',
            taxid, name) for taxid, name in list(zip(taxids, names))[:-1]
    ]



#!/usr/bin/env python

from __future__ import print_function

import sys
import argparse
import re

from dark.taxonomy import AccessionLineageFetcher, formatLineage

VERSION_REGEX = re.compile(r'\.\d+$')


def taxonomyInfo(accession, db, namesOnly, separator):
    """
    Convert the lineage information for C{accession} into a formatted string.

    @param accession: The C{str} accession number to look up.
    @param db: The sqlite3 taxonomy database to consult.
    @param namesOnly: If C{True} only print taxonomic names.
    @param separator: The C{str} separator to put between fields. If C{None}
        return space-padded aligned columns.
    @return: A formatted C{str} for printing or C{None} if C{accession} is
        not found in the taxonomy database.
    """
    lineage = db.lineage(accession)

    if lineage is None and VERSION_REGEX.search(accession) is None:
        # Try adding a version number.
        lineage = db.lineage(accession + '.1')

    if lineage:
        return formatLineage(lineage, namesOnly, separator)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Print taxonomy information for accession numbers')

    parser.add_argument(
        'accessions', nargs='*',
        help=('The accession number(s) to print taxonomy information for. If '
              'not given, accession numbers are read from standard input, one '
              'per line.'))

    parser.add_argument(
        '--separator',
        help=('The string to separate output columns with. Default is to '
              'print aligned output padded with multiple spaces (or a TAB if '
              '--namesOnly is used).'))

    parser.add_argument(
        '--database', required=True,
        help=('The file holding the sqlite3 taxonomy database. See '
              'https://github.com/acorg/ncbi-taxonomy-database for how to '
              'build one.'))

    parser.add_argument(
        '--namesOnly', default=False, action='store_true',
        help='If specified, only print the taxonomic names.')

    parser.add_argument(
        '--printAccession', default=False, action='store_true',
        help='If specified, also print the accession number.')

    args = parser.parse_args()

    db = AccessionLineageFetcher(args.database)

    if args.accessions:
        accessions = args.accessions
    else:
        accessions = (line[:-1] for line in sys.stdin)

    separator = args.separator

    if args.namesOnly and separator is None:
        separator = '\t'

    printAccession = args.printAccession

    for accession in accessions:
        if printAccession:
            print(accession + ':')
        result = taxonomyInfo(accession, db, args.namesOnly, separator)
        if result:
            print(result)
        else:
            print('%r was not found in the taxonomy database' % accession,
                  file=sys.stderr)



from unittest import TestCase
import six
import sqlite3

from dark.taxonomy import (
    LineageFetcher, AccessionLineageFetcher, isRetrovirus, isRNAVirus)


class FakeCursor(object):
    def __init__(self, results):
        self._results = results
        self._index = -1

    def execute(self, p):
        pass

    def fetchone(self):
        self._index += 1
        return self._results[self._index]

    def close(self):
        pass


class FakeDbConnection(object):
    def __init__(self, results):
        self._results = results
        self.open = True

    def cursor(self):
        return FakeCursor(self._results)

    def close(self):
        self.open = False


class TestLineageFetcher(TestCase):
    """
    Test LineageFetcher class.
    """
    def testGetTaxonomy(self):
        """
        Test if the LineageFetcher class works properly.
        """
        title = 'gi|5|gb|EU375804.1| Merkel cell polyomavirus'

        db = FakeDbConnection([
            [15], ['Merkel cell polyomavirus'],
            [4], ['Polyomavirus'],
            [3], ['dsDNA viruses'],
            [2], ['Vira'],
            [1],
        ])
        cursor = db.cursor()

        lineageFetcher = LineageFetcher(db=db, cursor=cursor)

        lineage = lineageFetcher.lineage(title)
        self.assertEqual(
            [
                (15, 'Merkel cell polyomavirus'),
                (4, 'Polyomavirus'),
                (3, 'dsDNA viruses'),
                (2, 'Vira'),
            ],
            lineage)


class TestAccessionLineageFetcher(TestCase):
    """
    Test the AccessionLineageFetcher class.
    """

    def _makeFetcher(self, data):
        """
        Make a taxonomy database and put it into an AccessionLineageFetcher
        instance.

        @param data: A C{dict} with 'taxids', 'names', and 'nodes' keys,
            each of which is a C{list} of C{tuple}s containing values to
            insert into the respective tables.
        @return: A C{AccessionLineageFetcher} instance.
        """
        db = sqlite3.connect(':memory:')
        cursor = db.cursor()

        # The tables created here must be identical to those expected by
        # the AccessionLineageFetcher class. See also
        # https://github.com/acorg/ncbi-taxonomy-database
        cursor.executescript('''
            CREATE TABLE nodes (
                taxid INTEGER NOT NULL,
                parent_taxid INTEGER NOT NULL,
                rank VARCHAR NOT NULL
            );

            CREATE TABLE accession_taxid (
                accession VARCHAR UNIQUE PRIMARY KEY,
                taxid INTEGER NOT NULL
            );

            CREATE TABLE names (
                taxid INTEGER NOT NULL,
                name VARCHAR NOT NULL
            );

            CREATE INDEX nodes_idx ON nodes(taxid);
            CREATE INDEX accession_idx ON accession_taxid(accession);
            CREATE INDEX name_idx ON names(taxid);
        ''')

        execute = cursor.execute

        for accession, taxid in data.get('taxids', []):
            execute('INSERT INTO accession_taxid VALUES (?, ?)',
                    (accession, taxid))

        for taxid, name in data.get('names', []):
            execute('INSERT INTO names VALUES (?, ?)',
                    (taxid, name))

        for taxid, parentTaxid, rank in data.get('nodes', []):
            execute('INSERT INTO nodes VALUES (?, ?, ?)',
                    (taxid, parentTaxid, rank))

        return AccessionLineageFetcher(db)

    def testUnknownTaxid(self):
        """
        If a taxonomy id is not present in the accession_taxid table, the
        lineage fetcher must return C{None}.
        """
        fetcher = self._makeFetcher({})
        self.assertIsNone(fetcher.lineage('DQ011818.1'))
        fetcher.close()

    def testReusingClosedFetcher(self):
        """
        Trying to re-use a closed fetcher must result in an AttributeError
        being raised.
        """
        fetcher = self._makeFetcher({})
        fetcher.close()
        error = "^'NoneType' object has no attribute 'cursor'$"
        six.assertRaisesRegex(self, AttributeError, error, fetcher.lineage,
                              'DQ011818.1')

    def testTaxidNotInNamesTable(self):
        """
        If a taxonomy id is not present in the names table, a ValueError must
        be raised.
        """
        fetcher = self._makeFetcher({
            'taxids': (
                ('DQ011818.1', 500),
            ),
        })
        error = '^Could not find taxonomy id 500 in names table$'
        six.assertRaisesRegex(self, ValueError, error, fetcher.lineage,
                              'DQ011818.1')
        fetcher.close()

    def testTaxidNotInNodesTable(self):
        """
        If a taxonomy id is not present in the nodes table, a ValueError must
        be raised.
        """
        fetcher = self._makeFetcher({
            'taxids': (
                ('DQ011818.1', 500),
            ),
            'names': (
                (500, 'Porcine endogenous retrovirus A'),
            ),
        })
        error = '^Could not find taxonomy id 500 in nodes table$'
        six.assertRaisesRegex(self, ValueError, error, fetcher.lineage,
                              'DQ011818.1')
        fetcher.close()

    def testFullLookup(self):
        """
        The full taxonomy of an accession number must be retrievable.
        """
        fetcher = self._makeFetcher({
            'taxids': (
                ('DQ011818.1', 500),
            ),
            'names': (
                (500, 'Porcine endogenous retrovirus A'),
                (400, 'Retroviruses'),
                (300, 'Viruses'),
            ),
            'nodes': (
                (500, 400, 'species'),
                (400, 300, 'genus'),
                (300, 1, 'realm')
            ),
        })

        self.assertEqual(
            (
                (500, 'Porcine endogenous retrovirus A', 'species'),
                (400, 'Retroviruses', 'genus'),
                (300, 'Viruses', 'realm'),
            ),
            fetcher.lineage('DQ011818.1'))
        fetcher.close()


class TestIsRetrovirus(TestCase):
    """
    Test the isRetrovirus function.
    """
    def testYes(self):
        """
        A retrovirus must be detected.
        """
        self.assertTrue(
            isRetrovirus((
                (11676, 'Human immunodeficiency virus 1', 'species'),
                (11646, 'Lentivirus', 'genus'),
                (327045, 'Orthoretrovirinae', 'subfamily'),
                (11632, 'Retroviridae', 'family'),
                (2169561, 'Ortervirales', 'order'),
                (10239, 'Viruses', 'superkingdom'),
            )))

    def testNo(self):
        """
        A non-retrovirus must be detected.
        """
        self.assertFalse(
            isRetrovirus((
                (10310, 'Human alphaherpesvirus 2', 'species'),
                (10294, 'Simplexvirus', 'genus'),
                (10293, 'Alphaherpesvirinae', 'subfamily'),
                (10292, 'Herpesviridae', 'family'),
                (548681, 'Herpesvirales', 'order'),
                (10239, 'Viruses', 'superkingdom'),
            )))


class TestIsRNAVirus(TestCase):
    """
    Test the isRNAVirus function.
    """
    def testByRealm(self):
        """
        Anything in the Riboviria realm must be classified as an RNA virus.
        """
        self.assertTrue(
            isRNAVirus((
                (11234, 'Measles morbillivirus', 'species'),
                (11229, 'Morbillivirus', 'genus'),
                (2560076, 'Orthoparamyxovirinae', 'subfamily'),
                (11158, 'Paramyxoviridae', 'family'),
                (11157, 'Mononegavirales', 'order'),
                (2497574, 'Monjiviricetes', 'class'),
                (2497570, 'Haploviricotina', 'subphylum'),
                (2497569, 'Negarnaviricota', 'phylum'),
                (2559587, 'Riboviria', 'realm'),
                (10239, 'Viruses', 'superkingdom'),
            )))

    def testHBV(self):
        """
        An HBV virus must not be classified as an RNA virus.
        """
        self.assertFalse(
            isRNAVirus((
                (1508712, 'Tent-making bat hepatitis B virus', 'species'),
                (10405, 'Orthohepadnavirus', 'genus'),
                (10404, 'Hepadnaviridae', 'family'),
                (10239, 'Viruses', 'superkingdom'),
            )))
